from __future__ import annotations

import abc
import json
import os
from configparser import ConfigParser, NoOptionError
from pathlib import Path
from typing import Any, Dict, List, Optional, Callable

import requests

from kttool.logger import color_green, log, log_cyan, log_red
from kttool.utils import (
    HEADERS, KATTIS_RC_URL, MAP_TEMPLATE_TO_PLANG, PLanguage, ask_with_default
)


class ConfigError(Exception):
    pass


def require_login(fn: Callable) -> Callable:
    def inner(self: 'Action', *args, **kwargs):
        did_login = False
        if not self.is_logged_in:
            did_login = True
            self.login()
        ret = fn(self, *args, **kwargs)
        if did_login:
            # poor man way to "timeout" the login attempt and invalidate the cookies
            self.is_logged_in = False
        return ret

    return inner


class Action(abc.ABC):
    ''' 
    Base class for handle general command.
    Handle loading up .kattisrc config file
    '''
    REQUIRED_CONFIG: bool = False

    _PRINTED_OUT_USERNAME: bool = False

    cwd: Path
    config_path: Path
    cfg: None | ConfigParser
    cookies: Any
    kt_config: Path

    file_name: None | Path
    lang: None | str
    pre_script: None | str
    script: None | str
    post_script: None | str
    is_logged_in: bool

    __slots__ = 'cwd', 'config_path', 'cfg', 'cookies', 'kt_config', 'file_name', 'lang', \
        'pre_script', 'script', 'post_script', 'is_logged_in'

    def __init__(self, *, cwd: None | Path = None):
        self.config_path = Path.home() / '.kattisrc'  # kattis config file
        self.kt_config = Path.home() / '.ktconfig'  # kt tool file
        self.cfg = None
        self.cwd = cwd or Path().absolute()
        self.cookies = None

        self.file_name = None
        self.lang = None
        self.pre_script = None
        self.script = None
        self.post_script = None
        self.is_logged_in = False

    def get_url(self, option: str, default: str = '') -> str:
        """ Get appropriate urls from kattisrc file

        Parameters
        ----------
        option : str
            parameter to get from katticrc config file
        default : str, optional
            fallback value if option is not present, by default ''

        Returns
        -------
        str
            Full url path to the required attr
        """
        if self.cfg.has_option('kattis', option):
            return self.cfg.get('kattis', option)

        kattis_host = self.cfg.get('kattis', 'hostname')
        return f'https://{kattis_host}/{default}'

    def read_config_from_file(self) -> None:
        """ kttool deals with 2 config files:
        - kattisrc: provided by official kattis website, provide domain name and general urls
        - ktconfig: handle templates by kttool

        Raises
        ------
        RuntimeError
            If config file is invalid
        ConfigError
            If config file is invalid
        """

        # Initialize ktconfig file if file doesnt exist
        if not self.kt_config.is_file():
            with open(self.kt_config, 'w') as f:
                f.write('{}\n')

        self.cfg = ConfigParser()
        if not self.config_path.is_file():
            raise RuntimeError(
                f'No valid config file at {self.config_path}. '
                f'Please download it at {KATTIS_RC_URL}'
            )

        self.cfg.read(self.config_path)
        username = self.cfg.get('user', 'username')
        password = token = None
        try:
            password = self.cfg.get('user', 'password')
        except NoOptionError:
            pass
        try:
            token = self.cfg.get('user', 'token')
        except NoOptionError:
            pass
        if password is None and token is None:
            raise ConfigError(
                '''\
        Your .kattisrc file appears corrupted. It must provide a token (or a
        KATTIS password).
        Please download a new .kattisrc file'''
            )

        if Action._PRINTED_OUT_USERNAME:
            return
        log(f'Username: {color_green(username)}')
        Action._PRINTED_OUT_USERNAME = True

    def login(self) -> None:
        """ Try to login and obtain cookies from succesful signin

        Raises
        ------
        RuntimeError
            If login fails
        """

        username = self.cfg.get('user', 'username')
        password = token = ''
        try:
            password = self.cfg.get('user', 'password')
        except NoOptionError:
            pass
        try:
            token = self.cfg.get('user', 'token')
        except NoOptionError:
            pass
        login_url = self.get_url('loginurl', 'login')
        login_args = {'user': username, 'script': 'true'}
        if password:
            login_args['password'] = password
        if token:
            login_args['token'] = token
        login_reply = self._request_post(login_url, data=login_args)

        if login_reply.status_code != 200:
            if login_reply.status_code == 403:
                err = 'Incorrect username or password/token (403)'
            elif login_reply.status_code == 404:
                err = 'Incorrect login URL (404)'
            else:
                err = f'Status code: {login_reply.status_code}'
            raise RuntimeError(f'Login failed. {err}')
        self.cookies = login_reply.cookies
        self.is_logged_in = True

    def _get_problem_id(self) -> str:
        # Assuming user is in the folder with the name of the problem id
        return self.cwd.name

    def _request_get(self, uri: str) -> requests.Response:
        return requests.get(uri, cookies=self.cookies, headers=HEADERS)

    def _request_post(self, uri: str, *args, **kwargs) -> requests.Response:
        return requests.post(
            uri, *args, **kwargs, cookies=self.cookies, headers=HEADERS
        )

    def get_problem_url(self, supplied_id: None | str = None) -> str:
        domain = f"https://{self.get_url('hostname')}"
        problem_id = supplied_id or self._get_problem_id()
        return os.path.join(domain, 'problems', problem_id)

    def _detect_code_files(self) -> bool:
        """ Confirm the executable file if there is multiple files that are runnable in current folder

        Raises
        ------
        RuntimeError
            If no executable code detected
        """
        acceptable_file_ext: Dict[str, PLanguage] = {}
        opt = 0
        i = 0
        existed_templates = self.load_kt_config()

        for k in existed_templates.keys():
            acceptable_file_ext[MAP_TEMPLATE_TO_PLANG[k].extension
                               ] = MAP_TEMPLATE_TO_PLANG[k]

        files = [x for x in self.cwd.iterdir() if x.is_file()]
        runnable_files: List[Path] = []
        for f in files:
            if f.suffix[1:] in acceptable_file_ext:
                runnable_files.append(f)

        if not runnable_files:
            log_red('No executable code files detected')
            return False

        if len(runnable_files) > 1:
            log_cyan('Choose a file to run')
            for i in range(len(runnable_files)):
                log(f'  {i}: {runnable_files[i].file_name}')
            opt = int(ask_with_default('Choose a file to run', default_val='0'))
            if not (0 <= opt < len(runnable_files)):
                log_red('Invalid option chosen')
                return False

        self.file_name = runnable_files[opt]
        alias = acceptable_file_ext[runnable_files[opt].suffix[1:]].alias
        self.lang = acceptable_file_ext[runnable_files[opt].suffix[1:]
                                       ].full_name

        file_name = self.file_name.stem
        self.pre_script = existed_templates.get(alias,
                                                {}).get('pre_script').replace(
                                                    '$%file%$', file_name
                                                )
        self.script = existed_templates.get(alias, {}).get('script').replace(
            '$%file%$', file_name
        )
        self.post_script = existed_templates.get(alias,
                                                 {}).get('post_script').replace(
                                                     '$%file%$', file_name
                                                 )
        return True

    def load_kt_config(self) -> dict:
        if not self.kt_config.is_file():
            with open(self.kt_config, 'w+') as f:
                json.dump({}, f)

        try:
            with open(self.kt_config) as f:
                return json.load(f)
        except:
            log_red('kattis config maybe corrupted, resetting..')

        with open(self.kt_config, 'w+') as f:
            json.dump({}, f)
        return {}

    @abc.abstractmethod
    def _act(self) -> None:
        raise NotImplementedError()

    def act(self) -> None:
        """Main entry to execute the command
        """
        if self.REQUIRED_CONFIG:
            self.read_config_from_file()
        self._act()



from __future__ import annotations

import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional
from typing_extensions import final
import bs4
from bs4 import BeautifulSoup

from ..base import Action, require_login
from ..logger import log, log_green, log_red
from ..utils import MAP_TEMPLATE_TO_PLANG

__all__ = ['Gen']


@dataclass(frozen=True)
class _SampleData:
    problem_id: str
    is_input: bool
    sample_id: str
    data: str


@final
class Gen(Action):
    """Usage: kt gen <problem_id>

    Generate a folder which contains the template code (for the default language of choice) and the sample inputs and outputs associated with the 
    problem id. 

    Options
    --------
    problem_id: Kattis problem id
    """
    REQUIRED_CONFIG = True

    problem_id: str
    __slots__ = 'problem_id'

    def __init__(self, problem_id: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.problem_id = self._parse_problem_id(problem_id)

    @staticmethod
    def _write_sample(sample_data: _SampleData, root_path: Path) -> None:
        """ Write input/output sample to file. This is used for multiprocess pool to generate input/output files
        Parameters
        ----------
        sample_data : SampleData
            a tuple representing index, string data, problem id and a boolean declaring whether current
        file is input (False if the file is output)
        """
        file_name_prefix = 'in' if sample_data.is_input else 'ans'
        file_name = root_path / f'{sample_data.problem_id}/{file_name_prefix}{sample_data.sample_id}.txt'

        with open(file_name, 'w+') as f:
            f.write(sample_data.data)

    @staticmethod
    def _parse_problem_id(raw_input: str) -> Optional[str]:
        def _is_url(s: str) -> bool:
            return 'open.kattis.com' in s

        def _parse_problem_id_from_url(url_like: str) -> Optional[str]:
            parts = url_like.split('/')
            if len(parts) == 0:
                return None
            return parts[-1]

        return raw_input if not _is_url(
            raw_input
        ) else _parse_problem_id_from_url(raw_input)

    def _parse_sample_data(self, data: bs4.ResultSet) -> List[_SampleData]:
        sample_data = []
        if data is None:
            return []
        for i in range(len(data)):
            if i & 1:
                sample_data.append(
                    _SampleData(
                        sample_id=i // 2 + 1,
                        data=data[i].text,
                        problem_id=self.problem_id,
                        is_input=False
                    )
                )
            else:
                sample_data.append(
                    _SampleData(
                        sample_id=i // 2 + 1,
                        data=data[i].text,
                        problem_id=self.problem_id,
                        is_input=True
                    )
                )
        return sample_data

    def _parse_samples_url(self, url: str) -> None | bs4.ResultSet:
        page = self._request_get(url)
        soup = BeautifulSoup(page.content, 'html.parser')
        ret: None | bs4.ResultSet = None
        for sample in soup.find_all('table', class_='sample'):
            if ret is None:
                ret = sample.find_all('pre')
            else:
                ret += sample.find_all('pre')
        return ret

    @require_login
    def _generate_samples(self) -> None:
        """ Generate sample input file for `self.problem_id`
        The basic flow is to scrape the problem task page and retrieve the relevent fields
        Generate the sample files to problem id folder
        For example, if the problem id is distinctivecharacter, `kt gen` will
        - Generate a folder called distinctivecharacter
        - Get sample intput, output from problem task page and generate to distinctivecharacter/, in this
        example there will be 4 files generated
        + distinctivecharacter/in1.txt
        + distinctivecharacter/ans1.txt
        + distinctivecharacter/in2.txt
        + distinctivecharacter/ans2.txt
        - Generate a template file (distinctivecharacter.cpp) if a template file is provided in the .ktconfig file
        """
        url = self.get_problem_url()
        data = self._parse_samples_url(url)
        sample_data = self._parse_sample_data(data)

        assert len(data) % 2 == 0, 'Internal error: Number of sample input '\
            f' is not equal to number of sample output data={data!r}'

        for sample in sample_data:
            self._write_sample(sample, self.cwd)

        log_green(
            f'Generate {len(sample_data) // 2} sample(s) to {self.problem_id}'
        )

    def _get_problem_id(self) -> str:
        return self.problem_id

    def _generate_template_file(self):
        """_summary_
        """
        template_file = {}
        if not self.kt_config.is_file():
            log_red(
                '.ktconfig file has not been set up so no template was generated. '
                'Please use `kt config` to set up a template file'
            )
            return

        template_file = self.load_kt_config()

        for k, template in template_file.items():
            if template.get('default', None):
                template_file_location = template.get('path')
                if not template_file_location or not Path(
                    template_file_location
                ).is_file():
                    continue
                target_location = self.cwd / f'{self.problem_id}/{self.problem_id}.{MAP_TEMPLATE_TO_PLANG[k].extension}'
                target_location.parent.mkdir(parents=True, exist_ok=True)
                shutil.copyfile(template_file_location, target_location)
                log_green('Template file has been generated')
                return
        log_red(f'No default template detected in {self.kt_config}')

    def _act(self) -> None:
        if self.problem_id is None:
            log_red(f'Unable to parse problem id')
            return

        log(f'Problem is {self.problem_id}')
        problem_dir = self.cwd / self.problem_id
        problem_dir.mkdir(parents=True, exist_ok=True)
        self._generate_samples()
        self._generate_template_file()



from __future__ import annotations
from ast import List
from dataclasses import dataclass
from pathlib import Path
import random
from .gen import Gen
from ..base import Action
from bs4 import BeautifulSoup
from typing_extensions import final

__all__ = ['Surprise']

DifficultyFixed = float
DifficultyRange = (float, float)


@final
@dataclass
class KattisProblem:
    id: str
    difficulty: DifficultyFixed | DifficultyRange


@final
class Surprise(Action):
    """Usage: kt surprise [lower_bound] [upperBound]

    Randomly retrieve a problem from Kattis whose difficulty belongs to the range
    Naturally lower_bound has to be <= upperbound

    Options
    --------
    lower_bound: lower range bound to randomize. Default is 0
    upper_bound: lower range bound to randomize. Default is 10
    """
    REQUIRED_CONFIG = True
    _FIRST_INDEX = 0
    _LAST_INDEX = 35
    __slots__ = '_easiest_difficulty', '_hardest_difficulty'

    def __init__(
        self,
        easiest_difficulty: float = 0.,
        hardest_difficulty: float = 10.,
        *,
        cwd: None | Path = None
    ):
        super().__init__(cwd=cwd)
        self._easiest_difficulty = float(easiest_difficulty)
        self._hardest_difficulty = float(hardest_difficulty)
        assert easiest_difficulty <= hardest_difficulty

    @staticmethod
    def _parse_difficulty(val: str) -> DifficultyFixed | DifficultyRange:
        try:
            return float(val)
        except:
            return tuple([float(x.strip()) for x in val.split('-')])

    @staticmethod
    def _parse_id(val: str) -> str:
        return val.split('/')[-1]

    def _get_random_list(self) -> List[KattisProblem]:
        ret = []
        page = self._request_get(
            f'https://{self.get_url("hostname")}/problems?page={random.randint(self._FIRST_INDEX - 1, self._LAST_INDEX)}&order=%2Bdifficulty_category'
        )
        soup = BeautifulSoup(page.content, 'html.parser')
        table = soup.find_all('table', class_='table2')[0].find_next('tbody')
        for row in table.find_all('tr'):
            ret.append(
                KattisProblem(
                    id=self._parse_id(row.find_next('a', href=True)['href']),
                    difficulty=self._parse_difficulty(
                        row.find_next('span', class_='difficulty_number').text
                    )
                )
            )
        return ret

    def _match(self, problem: KattisProblem) -> bool:
        if isinstance(problem.difficulty, DifficultyFixed):
            return self._easiest_difficulty <= problem.difficulty <= self._hardest_difficulty
        return self._easiest_difficulty <= problem.difficulty[0] \
            and problem.difficulty[1] <= self._hardest_difficulty

    def _problem_already_being_attempted(self, problem: KattisProblem) -> bool:
        if (self.cwd / problem.id).is_dir():
            return True
        return False

    def _act(self) -> None:
        problem: None | KattisProblem = None
        while problem is None:
            for a_problem in self._get_random_list():
                if not self._match(a_problem):
                    continue
                if self._problem_already_being_attempted(a_problem):
                    continue
                problem = a_problem
                break
        Gen(problem.id, cwd=self.cwd).act()



from pathlib import Path
import readline, glob, os, json
from ..logger import log, log_red
from ..base import Action
from ..logger import color_cyan, color_green, color_red, log_green
from ..utils import MAP_TEMPLATE_TO_PLANG, PLanguage, ask_with_default
from typing_extensions import final

__all__ = ['Config']


@final
class Config(Action):
    """Usage: kt config

    Create or update a template config for your code. This command should be ran the first time you used the tool or whenever you want to update your
    testing/language configs. Concretely, the config stores the information to:
    - Path to your template code (C++, py, etc) to copy whenever `kt gen` is ran
    - before_script that executed once before testing your code against the samples
    - script to run the code/binary
    - after_script that usually does the clean up
    """
    def add_template(self) -> None:
        question = 'Which template would you like to add:\n'
        selectable_lang = {}
        idx = 1
        existed_templates = {}
        options = {}

        log_green('Adapted from xalanq\'s cf tool')
        log(
            '''
Template will run 3 scripts in sequence when you run "kt test":
    - before_script   (execute once)
    - script          (execute the number of samples times)
    - after_script    (execute once)
You could set "before_script" or "after_script" to empty string, meaning not executing.
You have to run your program in "script" with standard input/output (no need to redirect).

You can insert some placeholders in your scripts. When execute a script,
cf will replace all placeholders by following rules:

$%path%$   Path to source file (Excluding $%full%$, e.g. "/home/user/")
$%full%$   Full name of source file (e.g. "a.cpp")
$%file%$   Name of source file (Excluding suffix, e.g. "a")
$%rand%$   Random string with 8 character (including "a-z" "0-9")
        '''
        )
        existed_templates = self.load_kt_config()
        for template_type, lang in MAP_TEMPLATE_TO_PLANG.items():
            if template_type not in existed_templates:
                temp = f'{idx} ({lang.extension}): {lang.full_name}\n'
                question += temp
                selectable_lang[idx] = (template_type, lang)
                idx += 1

        res = input(question)
        ret = int(res)
        assert 1 <= ret < idx, 'Invalid input'
        selected_lang: PLanguage = selectable_lang[ret][1]

        def complete(text, state):
            return (glob.glob(os.path.expanduser(text) + '*') + [None])[state]

        readline.set_completer_delims(' \t\n;')
        readline.parse_and_bind("tab: complete")
        readline.set_completer(complete)
        options['path'] = os.path.expanduser(
            ask_with_default(
                'Path to template file: ',
                default_val=self._fetch_default_template_file(
                    selected_lang.extension
                )
            )
        )
        options['pre_script'] = ask_with_default(
            'Pre-script', selected_lang.pre_script
        )
        options['script'] = ask_with_default('Script', selected_lang.script)
        options['post_script'] = ask_with_default(
            'Post-script', selected_lang.post_script
        )
        options['default'] = False if existed_templates else True

        existed_templates[selected_lang.alias] = options
        with open(self.kt_config, 'w') as kt_config:
            json.dump(existed_templates, kt_config, indent=2)
        log_green(
            f'Yosh, your configuration has been saved to {self.kt_config}'
        )

    @staticmethod
    def _fetch_default_template_file(extension: str) -> str:
        template_file = Path(
            __file__
        ).parent.parent / 'default_templates' / f'template.{extension}'
        if template_file.is_file():
            return str(template_file.absolute())
        return ''

    def remove_template(self) -> None:
        ''' Remove a template from ktconfig file'''
        existed_templates = self.load_kt_config()
        if not existed_templates:
            log('No existing templates')
            return

        log(
            f'Which template would you like to {color_red("delete")} ? For eg cpp, cc, ...'
        )
        for k in existed_templates.keys():
            log(k)
        res = input()

        assert res in existed_templates, f'Invalid template chosen. Template {res} is not in ur config file'

        move_default = existed_templates[res]['default']
        existed_templates.pop(res, None)
        if existed_templates and move_default:  # move default to the first key of template
            existed_templates[next(iter(existed_templates))] = True
        with open(self.kt_config, 'w') as kt_config:
            json.dump(existed_templates, kt_config, indent=2)

    def update_default(self) -> None:
        default_key = ''
        existed_templates = self.load_kt_config()

        log(
            f'Which template would you like to gen as {color_cyan("default")} ? For eg cpp, cc, ...'
        )

        for k, v in existed_templates.items():
            log(f'{k} {color_green("(default)") if v["default"] else ""}')
            if v["default"]:
                default_key = k
        res = input()
        if res not in existed_templates:
            log_red(
                f'Invalid template chosen. Template {res} is not in ur config file'
            )
            return

        if default_key:
            existed_templates[default_key]["default"] = False
        existed_templates[res]["default"] = True
        with open(self.kt_config, 'w') as kt_config:
            json.dump(existed_templates, kt_config, indent=2)
        log_green('Yosh, your configuration has been saved')

    def _act(self) -> None:
        question = color_cyan('Select an option:\n')
        question += """1: Add a template
2: Remove a template
3: Select a default template
"""
        res = input(question)
        opt = int(res)
        if opt == 1:
            self.add_template()
        elif opt == 2:
            self.remove_template()
        elif opt == 3:
            self.update_default()
        else:
            raise ValueError(f'Invalid option. Valid only [1-3]')



from __future__ import annotations

from typing import List, Type
from .actions.gen import Gen
from .actions.test import Test
from .actions.submit import Submit
from .actions.config import Config
from .actions.open import Open
from .actions.version import Version
from .actions.update import Update
from .actions.surprise import Surprise
from .base import Action
from .logger import log, log_red

map_key_to_class = {
    'gen': Gen,
    'test': Test,
    'submit': Submit,
    'config': Config,
    'open': Open,
    'version': Version,
    'update': Update,
    'surprise': Surprise,
}

action_with_aliases = {
    **map_key_to_class,
    'g': Gen,
    't': Test,
    's': Submit,
    'c': Config,
    'o': Open,
    'v': Version,
    'u': Update,
    'r': Surprise,
    'random': Surprise,
}


def _is_help(args: List[str]) -> bool:
    return any(x in args for x in {'-h', '--help', 'help'})


def _print_help(klass: Type[Action]) -> None:
    log(klass.__doc__)


def arg_parse(args: List[str]) -> None | Action:
    ''' Generate an appropriate command class based on user command stirng '''
    if len(args) == 0:
        raise ValueError(f'No command provided to kt')
    if args[0] not in action_with_aliases:
        raise ValueError(
            f'First argument should be one of {list(map_key_to_class.keys())}'
        )
    klass = action_with_aliases[args[0]]
    if _is_help(args[1:]):
        return _print_help(klass)

    try:
        return klass(*args[1:])
    except:
        log_red('Invalid usage')
        return _print_help(klass)



import sys, os

__all__ = ['supports_color']


def __supports_color():
    """
    Returns True if the running system's terminal supports color, and False
    otherwise.
    """
    plat = sys.platform
    supported_platform = plat != 'Pocket PC' and (
        plat != 'win32' or 'ANSICON' in os.environ
    )
    # isatty is not always implemented, #6223.
    is_a_tty = hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()
    return supported_platform and is_a_tty


supports_color = __supports_color()



from __future__ import annotations
from pathlib import Path
import webbrowser
from typing_extensions import final
from ..base import Action
from ..logger import log

__all__ = ['Open']


@final
class Open(Action):
    """Usage: kt open [problem_id]

    Open the default browser with the link to the full problem statement on Kattis. If no problem id is provided, the problem id wil be deduced using the 
    current directionary name
    
    Options
    --------
    problem_id: Kattis problem id
    """
    REQUIRED_CONFIG = True

    __slots__ = '_problem_id'

    def __init__(
        self, problem_id: None | str = None, *, cwd: None | Path = None
    ):
        super().__init__(cwd=cwd)
        self._problem_id = problem_id

    def _act(self) -> None:
        problem_url = self.get_problem_url(self._problem_id)
        log(f'Openning {problem_url}')
        webbrowser.open(problem_url)



from ..version import version
from ..base import Action
from ..logger import color_cyan, log
from typing_extensions import final

__all__ = ['Version']


@final
class Version(Action):
    """Usage: kt version

    Print out the version of kttool
    """
    def _act(self) -> None:
        log(f'Current version: {color_cyan(version)}')



import json
import os
import re
import shlex
import subprocess
import tempfile
import time
from dataclasses import dataclass
from pathlib import Path
from typing import List, Tuple
from typing_extensions import final
import psutil
from ..base import Action
from ..logger import (
    color_cyan, color_green, color_red, log, log_cyan, log_green, log_red,
    strike_through
)
from ..utils import (launch_subprocess, make_list_equal)

__all__ = ['Test']

AC = "Accepted".ljust(13, " ")
WA = "Wrong Answer".ljust(13, " ")


@dataclass
class Sample:
    index: int
    input_file: Path
    output_file: Path


@final
class Test(Action):
    """Usage: kt test

    Run the set of scripts to compile and test the runable code file. 
    - before_script that executed once before testing your code against the samples
    - script to run the code/binary
    - after_script that usually does the clean up
    """

    REQUIRED_CONFIG = True

    def _compare_samples(self, samples: List[Sample]) -> None:
        rusage_denom = 1 << 20

        for sample in samples:
            is_ac = True
            actual = []
            expected = []
            diff = []
            try:
                with open(sample.output_file, 'r') as f:
                    expected = [l.strip(" \n") for l in f.readlines()]
                with open(sample.input_file, 'rb') as f:
                    raw_input_fixture = f.read()
                p = launch_subprocess(
                    shlex.split(f'{self.script} -'),
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    shell=False,
                    preexec_fn=os.setsid
                )
                proc = psutil.Process(p.pid)
                mem_used = proc.memory_info().rss / rusage_denom
                start_time = time.perf_counter()
                raw_output = p.communicate(raw_input_fixture)[0].decode()
                p.wait()
                taken = time.perf_counter() - start_time
                actual = [z.strip(" \n") for z in raw_output.split('\n')]
                make_list_equal(actual, expected)

                for i in range(len(expected)):
                    ''' 
                    Compare the values line by line
                    For each line, compare the values from left to right. 
                    '''
                    ith_line_exp = [z for z in expected[i].split(' ')]
                    ith_line_actual = [z for z in actual[i].split(' ')]

                    make_list_equal(ith_line_exp, ith_line_actual)
                    current_diff = ''
                    for j in range(len(ith_line_exp)):
                        lhs = ith_line_exp[j]
                        rhs = ith_line_actual[j]
                        is_good, now_diff = self._compare_entity(rhs, lhs)
                        is_ac &= is_good
                        current_diff += now_diff

                    diff.append(current_diff)

                if is_ac:
                    log_green(
                        f'Test Case #{sample.index}: {AC} ... {taken:.3f} s   {mem_used:.2f} M'
                    )
                else:
                    log_red(
                        f'Test Case #{sample.index}: {WA} ... {taken:.3f} s   {mem_used:.2f} M'
                    )
                    log_cyan('--- Input ---')
                    log(raw_input_fixture.decode())
                    log_cyan('--- Diff ---')
                    for i in range(len(diff)):
                        log(diff[i])

            except subprocess.CalledProcessError as e:
                log_red(f'Test case #{sample.index}: Runtime Error {e!r}')
            except Exception as e:
                self._record_unexpected_exception(sample, e)

    @staticmethod
    def _record_unexpected_exception(sample: Sample, ex: Exception) -> None:
        import traceback
        p = Path(tempfile.gettempdir())
        p.mkdir(parents=True, exist_ok=True)
        tmp_file = p / 'kt_test.log'
        with open(tmp_file, 'w+') as f:
            f.write(traceback.format_exc())
        log_red(
            f'Test case #{sample.index}: Internal Error {ex!r}. More info at {tmp_file}'
        )

    def _gather_samples(self) -> List[Sample]:
        input_files = [
            x for x in self.cwd.iterdir()
            if x.is_file() and x.stem.startswith('in')
        ]
        output_files = [
            x for x in self.cwd.iterdir()
            if x.is_file() and x.stem.startswith('ans')
        ]
        usable_samples: List[Sample] = []

        in_pattern = re.compile("in(\d+).txt")
        ans_pattern = re.compile("ans(\d+).txt")
        for input_file in input_files:
            idx = int(in_pattern.search(input_file.name).group(1))
            for output_file in output_files:
                if idx == int(ans_pattern.search(output_file.name).group(1)):
                    usable_samples.append(
                        Sample(
                            index=idx,
                            input_file=Path(input_file).absolute(),
                            output_file=Path(output_file).absolute()
                        )
                    )
                    break
        # run test from ascending number of file index
        return sorted(usable_samples, key=lambda x: x.index)

    def _act(self) -> None:
        """ Run the executable file against sample input and output files present in the folder
        The sample files will only be recognized if the conditions hold:
        - Naming style should be in{idx}.txt and ans{txt}.txt
        - for in{idx}.txt, there must exist a ans{idx}.txt with the same `idx`
        """
        if not self._detect_code_files():
            return

        # Get sample files that match the condition
        usable_samples = self._gather_samples()
        # run test
        log(f'Problem ID : {color_cyan(self._get_problem_id())}')
        log(f'Lanuage    : {self.lang}')
        if self.pre_script:
            log_cyan(f'running {self.pre_script}')
            subprocess.check_call(shlex.split(self.pre_script))

        self._compare_samples(usable_samples)

        if self.post_script:
            log_cyan(f'running {self.post_script}')
            subprocess.check_call(shlex.split(self.post_script))

    @staticmethod
    def _compare_entity(lhs: str, rhs: str) -> Tuple[bool, str]:
        if lhs == rhs:
            return True, f'{lhs} '
        return False, f'{color_red(strike_through(lhs))}{color_green(rhs)} '


from setuptools import setup, find_packages
from kttool.version import version
import pathlib, os
from distutils.command.install import INSTALL_SCHEMES

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

for scheme in INSTALL_SCHEMES.values():
    scheme['data'] = scheme['purelib']

required_files = ['kttool/VERSION', 'LICENSE', 'requirements.txt']
for p in (pathlib.Path('kttool') / 'default_templates').iterdir():
    required_files += [os.path.relpath(p, pathlib.Path(__file__).parent)]

with open('requirements.txt') as f:
    deps = f.read().splitlines()

setup(
    name='kttool',
    version=version,
    description="Kattis command line tool",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/heiseish/kt",
    author="Dao Truong Giang",
    author_email="dtrnggiang@gmail.com",
    license="MIT",
    classifiers=[
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(),
    include_package_data=True,
    package_data={'kttool': required_files},
    data_files=[('kttool', required_files)],
    install_requires=deps,
    scripts=['kt']
)
